from django.http import HttpRequest, HttpResponse
from django.utils.functional import SimpleLazyObject

import pytest

from maykin_2fa.middleware import OTPMiddleware


@pytest.fixture()
def user_request(request, client, rf, user) -> HttpRequest:
    marker = request.node.get_closest_marker("user_request_auth_backend")
    backend = None if not marker else marker.args[0]
    # sets the backend and session
    client.force_login(user, backend=backend)

    # create a standalone request instance, by copying over the session and mimicking
    # the `django.contrib.auth.middleware.AuthenticationMiddleware` middleware
    request = rf.get("/irrelevant")
    request.session = client.session
    request.user = SimpleLazyObject(lambda: user)
    return request


@pytest.mark.user_request_auth_backend("django.contrib.auth.backends.ModelBackend")
def test_authenticated_but_not_verified(user_request, settings):
    """
    Test that a user who logs in is not verified by default.
    """
    settings.MAYKIN_2FA_ALLOW_MFA_BYPASS_BACKENDS = []
    middleware = OTPMiddleware(lambda req: HttpResponse())

    middleware(user_request)

    # standard Django behaviour
    user = user_request.user
    assert user.is_authenticated
    assert user.backend == "django.contrib.auth.backends.ModelBackend"
    # OTP middleware adds the `is_verified` callable
    assert hasattr(user, "is_verified")
    assert callable(user.is_verified)
    # we didn't go through any 2FA flows, so the user is *NOT* verified
    assert user.is_verified() is False


@pytest.mark.user_request_auth_backend("testapp.backends.No2FAModelBackend")
def test_authenticated_and_2fa_verification_bypassed(user_request, settings):
    """
    Test that a user is "2FA-verified" when authenticated through a backend on the
    allow list.
    """
    settings.MAYKIN_2FA_ALLOW_MFA_BYPASS_BACKENDS = [
        "testapp.backends.No2FAModelBackend"
    ]
    middleware = OTPMiddleware(lambda req: HttpResponse())

    middleware(user_request)

    # standard Django behaviour
    user = user_request.user
    assert user.is_authenticated
    assert user.backend == "testapp.backends.No2FAModelBackend"
    # OTP middleware adds the `is_verified` callable
    assert hasattr(user, "is_verified")
    assert callable(user.is_verified)
    # didn't go through any 2FA flows, but the backend is on the allow list
    assert user.is_verified()



from django.core.exceptions import ImproperlyConfigured

from two_factor.admin import AdminSiteOTPRequired


class MFARequired(AdminSiteOTPRequired):
    def login(self, request, extra_context=None):
        """
        Disabled to enforce usage of the custom login views.
        """
        raise ImproperlyConfigured(
            "Ensure the maykin_2fa urls are included *before* the default "
            "admin.site.urls."
        )



import functools

from django.contrib.auth.signals import user_logged_in
from django.dispatch import receiver

from .middleware import is_verified


@receiver(user_logged_in, dispatch_uid="maykin_2fa.add_is_verified_method")
def add_is_verified_method(sender, request, user, **kwargs):
    """
    Add the method ``is_verified`` to the user instance.

    Normally this is added through middleware, but because django's ``login`` function
    assigns `request.user = instance`` directly from the DB, this method is lost.
    """
    user.is_verified = functools.partial(is_verified, user)
    if not hasattr(user, "otp_device"):
        user.otp_device = None



from io import StringIO

from django.core.management import call_command
from django.core.management.base import SystemCheckError

import pytest


@pytest.mark.urls("tests.bad_urls.missing_core")
def test_urlconf_check_core_missing():
    with pytest.raises(SystemCheckError, match="(maykin_2fa.E001)"):
        call_command("check")


@pytest.mark.urls("tests.bad_urls.missing_webauthn")
def test_urlconf_check_webauthn_missing():
    with pytest.raises(SystemCheckError, match="(maykin_2fa.E002)"):
        call_command("check")


@pytest.mark.urls("tests.bad_urls.missing_webauthn")
def test_urlconf_without_webauhn_ok(settings):
    from testapp.settings import INSTALLED_APPS

    settings.INSTALLED_APPS = [
        app for app in INSTALLED_APPS if app != "two_factor.plugins.webauthn"
    ]
    settings.SILENCED_SYSTEM_CHECKS = ["maykin_2fa.E005"]

    call_command("check", stdout=StringIO())


def test_bypass_backend_not_in_auth_backends(settings):
    settings.MAYKIN_2FA_ALLOW_MFA_BYPASS_BACKENDS = ["__unknown__"]
    settings.SILENCED_SYSTEM_CHECKS = ["maykin_2fa.E005"]
    stderr = StringIO()

    call_command("check", stderr=stderr)

    assert "(maykin_2fa.W001)" in stderr.getvalue()


def test_wrong_middleware(settings):
    settings.MIDDLEWARE = [
        "django.middleware.security.SecurityMiddleware",
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.middleware.common.CommonMiddleware",
        "django.middleware.csrf.CsrfViewMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "django_otp.middleware.OTPMiddleware",
        "django.contrib.messages.middleware.MessageMiddleware",
        "django.middleware.clickjacking.XFrameOptionsMiddleware",
    ]

    with pytest.raises(SystemCheckError, match="(maykin_2fa.E004)"):
        call_command("check")


def test_middleware_missing(settings):
    settings.MIDDLEWARE = [
        "django.middleware.security.SecurityMiddleware",
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.middleware.common.CommonMiddleware",
        "django.middleware.csrf.CsrfViewMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "django.contrib.messages.middleware.MessageMiddleware",
        "django.middleware.clickjacking.XFrameOptionsMiddleware",
    ]

    with pytest.raises(SystemCheckError, match="(maykin_2fa.E003)"):
        call_command("check")


def test_okay_configuration():
    # testapp.settings is a valid configuration
    call_command("check", stdout=StringIO())



import functools
from typing import TypeAlias

from django.conf import settings
from django.contrib.auth import BACKEND_SESSION_KEY
from django.contrib.auth.models import AbstractBaseUser, AnonymousUser
from django.http import HttpRequest

from django_otp.middleware import (
    OTPMiddleware as _OTPMiddleware,
    is_verified as otp_is_verified,
)

# probably a TypeVar(bound=...) makes more sense but we don't allow using generics here,
# so let's cover the most ground using a simple union.
AnyUser: TypeAlias = AbstractBaseUser | AnonymousUser


def is_verified(user: AbstractBaseUser) -> bool:
    """
    Modified version of :func:`django_otp.middleware.is_verified` to add bypass logic.

    This function may not be called with an :class:`AnonymousUser` instance!

    If the user backend that the user authenticated with is on the allow list, we do
    not perform the real OTP device check from the library, but just mark the user as
    verified.
    """
    # check our allow list for authentication backends
    backends_to_skip_verification_for = getattr(
        settings, "MAYKIN_2FA_ALLOW_MFA_BYPASS_BACKENDS", []
    )
    if (
        backends_to_skip_verification_for
        and user.is_authenticated
        and user.backend in backends_to_skip_verification_for  # type: ignore
    ):
        return True
    # fall back to default library behaviour
    return otp_is_verified(user)


class OTPMiddleware(_OTPMiddleware):
    """
    Substitute our own :func:`is_verified` check instead of the upstream one.

    This middleware *replaces* :class:`django_otp.middleware.OTPMiddleware` to add
    allow-list behaviour for certain authentication backends. This marks the user
    authentication as being verified even though the 2FA requirements in the project
    itself have been bypassen.

    This setup allows us to enforce 2FA when logging in with username + password, but
    be less strict when signing it via OIDC/other SSO solutions that already enforce
    MFA at another level outside of our scope.
    """

    def _verify_user(self, request: HttpRequest, user: AnyUser):
        # call the super but replace the `is_verified` callable
        user = super()._verify_user(request, user)
        # this is *not* persisted on the user object after authenticate
        user.backend = request.session.get(BACKEND_SESSION_KEY)
        user.is_verified = functools.partial(is_verified, user)  # type: ignore
        return user



from django.contrib import admin
from django.contrib.auth import logout
from django.shortcuts import resolve_url
from django.utils.translation import gettext_lazy as _

from two_factor.forms import TOTPDeviceForm
from two_factor.utils import default_device
from two_factor.views import (
    BackupTokensView as _BackupTokensView,
    LoginView as _LoginView,
    QRGeneratorView as _QRGeneratorView,
    SetupCompleteView as _SetupCompleteView,
    SetupView as _SetupView,
)


class AdminLoginView(_LoginView):
    template_name = "maykin_2fa/login.html"
    redirect_authenticated_user = False

    def get_redirect_url(self):
        # after succesful authentication, check if the user needs to set up 2FA. If MFA
        # was configured already, login flow takes care of the OTP step.
        user = self.request.user

        if user.is_authenticated and not user.is_verified():
            # if no device is set up, redirect to the setup.
            device = default_device(user)
            if device is None:
                return resolve_url("maykin_2fa:setup")

            # a device is configured, but wasn't used - this may have been an aborted
            # authentication process. Log the user out and have the go through the login
            # flow again.
            logout(self.request)
            return resolve_url("maykin_2fa:login")

        admin_index = resolve_url("admin:index")
        return super().get_redirect_url() or admin_index

    def get_context_data(self, form, **kwargs):
        context = super().get_context_data(form, **kwargs)
        context.update(
            {
                **admin.site.each_context(self.request),
                "title": _("Log in"),
                "subtitle": None,
                "app_path": self.request.get_full_path(),
            }
        )
        return context


class AdminSetupView(_SetupView):
    # TODO: update to our own templates/URLs
    success_url = "maykin_2fa:setup_complete"
    qrcode_url = "maykin_2fa:qr"
    template_name = "maykin_2fa/setup.html"

    def get_form(self, *args, **kwargs):
        form = super().get_form(*args, **kwargs)

        # patch the form input type to not have numeric input controls...
        # I checked that this does not mutate the entire class :)
        if isinstance(form, TOTPDeviceForm):
            form.fields["token"].widget.input_type = "text"

        return form

    def get_context_data(self, form, **kwargs):
        context = super().get_context_data(form, **kwargs)
        context.update(
            {
                **admin.site.each_context(self.request),
                "title": _("Set up MFA"),
                "subtitle": None,
                "app_path": self.request.get_full_path(),
                # Cancelling MFA setup is not optional.
                "cancel_url": None,
            }
        )
        return context


class BackupTokensView(_BackupTokensView):
    success_url = "maykin_2fa:backup_tokens"
    # TODO
    template_name = "two_factor/core/backup_tokens.html"


class SetupCompleteView(_SetupCompleteView):
    template_name = "maykin_2fa/setup_complete.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            {
                **admin.site.each_context(self.request),
                "title": _("MFA setup complete"),
                "subtitle": None,
                "app_path": self.request.get_full_path(),
            }
        )
        return context


class QRGeneratorView(_QRGeneratorView):
    pass



def monkeypatch_admin():
    """
    Monkeypatch the admin to enforce 2FA.

    2FA is enforced but can be bypassed by putting the relevant authentication
    backend(s) on the allow list. In development, you might want to set all backends
    as bypassed:

    .. code-block::

        MAYKIN_2FA_ALLOW_MFA_BYPASS_BACKENDS = AUTHENTICATION_BACKENDS

    Upstream documentation:
    https://django-two-factor-auth.readthedocs.io/en/stable/implementing.html#admin-site
    """
    from django.contrib import admin

    from .admin import MFARequired

    admin.site.__class__ = MFARequired



from django.apps import apps
from django.conf import settings
from django.contrib import admin
from django.core.checks import Error, Warning, register
from django.urls import NoReverseMatch, reverse

from .admin import MFARequired


@register()
def check_urlconf(app_configs, **kwargs):
    errors = []
    try:
        reverse("maykin_2fa:login")
    except NoReverseMatch:
        example_code = 'path("admin/", include((urlpatterns, "maykin_2fa"))),'
        errors.append(
            Error(
                "Broken URL config - could not resolve the 'maykin_2fa:login' URL.",
                hint=f"Try including `{example_code}` in your root `urls.py`.",
                id="maykin_2fa.E001",
            )
        )

    if apps.is_installed("two_factor.plugins.webauthn"):
        try:
            reverse("two_factor:webauthn:create_credential")
        except NoReverseMatch:
            example_code = (
                'path("admin/", include((webauthn_urlpatterns, "two_factor"))),'
            )
            errors.append(
                Error(
                    "Broken URL config - could not resolve the "
                    "'two_factor:webauthn:create_credential' URL.",
                    hint=f"Try including `{example_code}` in your root `urls.py`.",
                    id="maykin_2fa.E002",
                )
            )

    return errors


@register()
def check_authentication_backends(app_configs, **kwargs):
    bypass_backends = getattr(settings, "MAYKIN_2FA_ALLOW_MFA_BYPASS_BACKENDS", [])
    unknown_backends = [
        backend
        for backend in bypass_backends
        if backend not in settings.AUTHENTICATION_BACKENDS
    ]
    if not unknown_backends:
        return []

    return [
        Warning(
            "MAYKIN_2FA_ALLOW_MFA_BYPASS_BACKENDS contains backends not present in the "
            f"AUTHENTICATION_BACKENDS setting: {', '.join(unknown_backends)}",
            hint=(
                "Check for typos or add the backend(s) to "
                "settings.AUTHENTICATION_BACKENDS"
            ),
            id="maykin_2fa.W001",
        )
    ]


@register()
def check_middleware(app_configs, **kwargs):
    errors = []

    if "maykin_2fa.middleware.OTPMiddleware" not in settings.MIDDLEWARE:
        errors.append(
            Error(
                "`maykin_2fa.middleware.OTPMiddleware` is missing from the middleware.",
                hint="Add the maykin_2fa middleware instead of the django_otp one.",
                id="maykin_2fa.E003",
            )
        )

    if "django_otp.middleware.OTPMiddleware" in settings.MIDDLEWARE:
        errors.append(
            Error(
                "Found `django_otp.middleware.OTPMiddleware` in the middleware - this "
                "is obsolete.",
                hint=(
                    "Remove the django_otp middleware (instead, use the maykin_2fa "
                    "one)."
                ),
                id="maykin_2fa.E004",
            )
        )

    return errors


@register()
def check_admin_patched(app_configs, **kwargs):
    cls = admin.site.__class__
    if cls is MFARequired or issubclass(cls, MFARequired):
        return []

    return [
        Error(
            "The default admin site is not monkeypatched.",
            hint="Call 'maykin_2fa.monkeypatch_admin' in your root urls.py",
            id="maykin_2fa.E005",
        )
    ]



from django.urls import include, path

from .views import (
    AdminLoginView,
    AdminSetupView,
    BackupTokensView,
    QRGeneratorView,
    SetupCompleteView,
)

# See two_factor/urls.py for a reference of the (core) urls

urlpatterns = [
    path("login/", AdminLoginView.as_view(), name="login"),
    path("mfa/setup/", AdminSetupView.as_view(), name="setup"),
    path("mfa/qrcode/", QRGeneratorView.as_view(), name="qr"),
    path("mfa/setup/complete/", SetupCompleteView.as_view(), name="setup_complete"),
    path("mfa/backup/tokens/", BackupTokensView.as_view(), name="backup_tokens"),
]

webauthn_urlpatterns = [
    path(
        "mfa/webauthn/",
        include("two_factor.plugins.webauthn.urls", namespace="two_factor_webauthn"),
    ),
]
